import util.io.env

BASE_DIR_ENV_NAME = 'MEASUREMENTS_DIR'
BASE_DIR = util.io.env.load(BASE_DIR_ENV_NAME)

## earth
EARTH_RADIUS = 6371 * 10**3
MAX_SEA_DEPTH = 11 * 10**3

## mean
MEAN_MIN_MEASUREMENTS = 2

## deviation
DEVIATION_MIN_MEASUREMENTS = 3
DEVIATION_MIN_VALUE = 0.05

## correlation
CORRELATION_MIN_ABS_VALUE = 0.01
CORRELATION_MAX_ABS_VALUE = 0.99
CORRELATION_MIN_DIAG_VALUE_POSITIVE_DEFINITE_APPROXIMATION = 0.1
