from measurements.constants import DEVIATION_MIN_MEASUREMENTS, DEVIATION_MIN_VALUE, MEAN_MIN_MEASUREMENTS, CORRELATION_MIN_ABS_VALUE, CORRELATION_MAX_ABS_VALUE

T_RANGE = [0, 1]
X_RANGE = [0, 360]
Y_RANGE = [-90, 90]